//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#ifndef VKRecorderRoot_H_
#define VKRecorderRoot_H_

#include "VKRecorderBase.hh"
#include "VKVirtualDetector.hh"
#include "G4PrimaryVertex.hh"
#include "G4PrimaryParticle.hh"
#include "G4VPhysicalVolume.hh"

#include <TTree.h>
#include <TFile.h>


// forward declarations
class VKDetectorConstruction;
class VKRecorderRootMessenger;


class VKRecorderRoot : public VKRecorderBase {

    public:
    
        VKRecorderRoot(G4String name = "");
        virtual ~VKRecorderRoot();

        virtual void RecordBeginOfRun(const G4Run*);
        virtual void RecordEndOfRun(const G4Run*);
        virtual void RecordBeginOfEvent(const G4Event*);
        virtual void RecordEndOfEvent(const G4Event*);
        virtual void RecordTrack(const G4Track*);
        virtual void RecordStep(const G4Step*);
        
        // set name of run macro (contents will be dumped to Root file)
        inline void SetRunMacroFileName(G4String name) {fRunMacroFileName=name;}
        
        // write data to current Root file and close file
        void WriteToRootFileAndClose();
        
        // open new Root file
        inline void SetRootFilename(G4String filename) {fFilename=filename;}
        
        // set zero suppression ON/OFF
        inline void SetZeroSuppressionOn(G4bool b) {fZeroSuppression=b;}

        // is zero suppression ON/OFF?
        inline G4bool GetZeroSuppressionOn() const {return fZeroSuppression;}
        
        // get energy deposited in SIGNAL and VETO detectors
        inline G4double GetEDepSignal() const {return fEDepSignal;}
        inline G4double GetEDepVeto()   const {return fEDepVeto;}
        
        // get data from last electron stepping through FOCAL plane
        inline G4double GetXfp()     const {return fX_fp;}
        inline G4double GetYfp()     const {return fY_fp;}
        inline G4double GetZfp()     const {return fZ_fp;}
        inline G4double GetEkinfp()  const {return fEkin_fp;}
        inline G4double GetThetafp() const {return fTheta_fp;}
        inline G4double GetPhifp()   const {return fPhi_fp;}

        // get data from last electron stepping through MIDWAY plane
        inline G4double GetXmp()     const {return fX_mp;}
        inline G4double GetYmp()     const {return fY_mp;}
        inline G4double GetZmp()     const {return fZ_mp;}
        inline G4double GetEkinmp()  const {return fEkin_mp;}
        inline G4double GetThetamp() const {return fTheta_mp;}
        inline G4double GetPhimp()   const {return fPhi_mp;}

        // get data from last primary
        inline G4double GetXso()     const {return fX_so;}
        inline G4double GetYso()     const {return fY_so;}
        inline G4double GetZso()     const {return fZ_so;}
        inline G4double GetEkinso()  const {return fEkin_so;}
        inline G4double GetThetaso() const {return fTheta_so;}
        inline G4double GetPhiso()   const {return fPhi_so;}
        
    private:
    
        void CreateDetBranches(); // detection branches
        void CreateTraBranches(); // trajectory branches
        void WriteEventToTree();
        void ClearRootBranches();
        void CloseRootFile();
        void WriteMetaDataToRootFile();
        void OpenRootFile(G4String filename);
        
        VKRecorderRootMessenger * fRecorderRoot_msg;
        
        G4bool fRootOutputOn, fZeroSuppression;
        
        G4String fRunMacroFileName;
 
        G4String fFilename;
        G4bool fBetaOn, fVirtualOn;
        const VKDetectorConstruction *fDetector;
        const VKVirtualDetector *fFocalPlane, *fMidwayPlane;
        const G4VPhysicalVolume *fSignalPV, *fVetoPV;
        
        G4PrimaryVertex *fPrimVertex;
        G4PrimaryParticle *fPrimParticle;
        
        G4double fX_fp, fY_fp, fZ_fp, fEkin_fp, fTheta_fp, fPhi_fp;  // focal plane
        G4double fX_mp, fY_mp, fZ_mp, fEkin_mp, fTheta_mp, fPhi_mp;  // midway plane
        G4double fX_so, fY_so, fZ_so, fEkin_so, fTheta_so, fPhi_so;  // source
        
        G4double fEDepSignal;
        G4double fEDepVeto;

        // ROOT output
        TFile *fTFile;
	    TTree *fDetTree, *fTraTree;
        
        TFile *ftest;
        
        Float_t _eDepSignal;
        Float_t _eDepVeto;
        UInt_t  _photonsSignal;
        UInt_t  _photonsVeto;
        
        Float_t _Xfp, _Yfp, _Zfp, _Efp, _THfp, _PHIfp;
        Float_t _Xmp, _Ymp, _Zmp, _Emp, _THmp, _PHImp;
        Float_t _Xso, _Yso, _Zso, _Eso, _THso, _PHIso;
};

#endif
