//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
#include "globals.hh"

#include "VKBlinkers.hh"
#include "VKUtil.hh"

#include "G4LogicalSkinSurface.hh"
#include "G4LogicalBorderSurface.hh"
#include "G4SystemOfUnits.hh"
#include "G4Cons.hh"
#include "TMath.h"


VKBlinkers::VKBlinkers(G4RotationMatrix *pRot,
                             const G4ThreeVector &tlate,
                             G4LogicalVolume *pMother_log,
                             G4bool pMany,
                             G4int pCopyNo,
                             const VKDetectorConstruction* pDetConstr)

 // Pass info to the G4PVPlacement constructor
 : G4PVPlacement(pRot, 
                 tlate,
                 CreateTemporaryLogicalVolume(), // <-- see VKUtil.cpp
                 "blinkers", 
                 pMother_log, 
                 pMany, 
                 pCopyNo),
   fDetConstr(pDetConstr),              
   fMother_log(pMother_log)
{

    // default dimensions    
    SetDefaults();

    // List of materials
    fMaterials = VKMaterials::GetInstance();

    // Get materials
    G4Material* pmma  = fMaterials -> GetMaterial("G4_PLEXIGLASS");
    G4Material* labVacuum = fMaterials -> GetMaterial("LABVACUUM");
    assert(pmma != nullptr); // check that material exists
    assert(labVacuum != nullptr);
        
    // rotation matrices, translation and transformation
    G4RotationMatrix doNotRotate; 
    doNotRotate.rotateX(0.);
    G4double displ, startAngle, spanningAngle;
    G4ThreeVector translation; 
    G4Transform3D transform;


    // -------------------------------------------- //
    //             (Imaginary) container            //
    // -------------------------------------------- //
    
    // *** All other volumes should fit inside this volume !

    // solid
    fContainer_sol = new G4Tubs("blinkers_sol", fInnerDiameter/2, fOuterDiameter/2, 
                              fTotalLength/2, 0.*deg, 360.*deg);

    // logical
    fContainer_log = new G4LogicalVolume(fContainer_sol, labVacuum, "blinkers_log");
    
    
    // --------------------------------- //
    //            PLASTIC BLINKERS       //
    // --------------------------------- //

    // Tube
    G4Tubs* tube_sol = new G4Tubs("blinkersTube_sol",
                                        fInnerDiameter/2, 
                                        fOuterDiameter/2,
                                        (fTotalLength-fConeLength)/2,
                                        startAngle=0.*deg, 
                                        spanningAngle=360.*deg);
                 
    // Cone
    G4Cons* cone_sol = new G4Cons("blinkersCone_sol",
                                      fInnerDiameter/2,
                                      fOuterDiameter/2,
                                      fOuterDiameter/2,
                                      fOuterDiameter/2,
                                      fConeLength/2,
                                      startAngle=0.*deg, 
                                      spanningAngle=360.*deg);
                 
    // placement of cone relative to tube
    displ       = fTotalLength/2;
    translation = G4ThreeVector(0, 0, displ); 
    transform   = G4Transform3D(doNotRotate, translation);

    // union of cone and tube
    fBlinkers_sol = new G4UnionSolid("blinkers_sol", tube_sol, cone_sol, transform);

    // logical volume
    fBlinkers_log = new G4LogicalVolume(fBlinkers_sol, pmma, "blinkers_log");

    // placement  
    new G4PVPlacement(0, G4ThreeVector(0, 0, -fConeLength/2), fBlinkers_log, 
                      "blinkers", fContainer_log, false, 0, fCheckOverlaps);

    // visualization attributes
    VisAttributes();    

    // Set the logical volume
    this -> SetLogicalVolume(fContainer_log);
    
    // Properly align blinkers along z axis
    G4ThreeVector pos = tlate + G4ThreeVector(0, 0, fTotalLength/2);
    this -> SetTranslation(pos);
    
    // Check for overlaps
    if (fCheckOverlaps) this -> CheckOverlaps();
}


void VKBlinkers::SetDefaults(){

    fCheckOverlaps = fDetConstr -> GetCheckOverlaps();

    // dimensions
    fCutOffAngle   = 65*deg;
    fThickness     = 18.5*mm; 
    fInnerDiameter = fDetConstr -> GetIronEndcapHoleDiameter();
    
    // deduced dimensions
    fOuterDiameter = fInnerDiameter + 2*fThickness;
    fTotalLength   = 0.5*fOuterDiameter / TMath::Tan(fCutOffAngle/rad);
    fConeLength    = fTotalLength - 0.5*fInnerDiameter / TMath::Tan(fCutOffAngle/rad); 
}


void VKBlinkers::VisAttributes() {

    // Blinkers
    G4VisAttributes* blinkers_va = new G4VisAttributes(G4Colour(0.3, 0.3, 0.8)); // ?
    blinkers_va   -> SetVisibility(true);
    blinkers_va   -> SetForceSolid(true);
    fBlinkers_log -> SetVisAttributes(blinkers_va);
}
