//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
#include "VKDetectorConstruction.hh"
#include "VKDetectorMessenger.hh"
#include "VKMaterials.hh"
#include "VKMagnet.hh"
#include "VKBetaDetector.hh"
#include "VKCatcherFoil.hh"
#include "VKShield.hh"
#include "VKLeadBlock.hh"
#include "VKBlinkers.hh"
#include "VKVirtualDetector.hh"
#include "VKTabulatedField3D.hh"
#include "VKSiPMSensitiveDetector.hh"
#include "VKScintillatorSensitiveDetector.hh"

#include "G4SDManager.hh"
#include "G4RunManager.hh"
#include "G4GeometryManager.hh"
#include "G4UImanager.hh"
#include "G4FieldManager.hh"
#include "G4TransportationManager.hh"

#include "G4SolidStore.hh"
#include "G4LogicalVolumeStore.hh"
#include "G4PhysicalVolumeStore.hh"

#include "G4LogicalBorderSurface.hh"
#include "G4LogicalSkinSurface.hh"
#include "G4OpticalSurface.hh"
#include "G4LogicalVolume.hh"
#include "G4PVPlacement.hh"
#include "G4Material.hh"
#include "G4MaterialTable.hh"
#include "G4VisAttributes.hh"

#include "G4Box.hh"
#include "G4Tubs.hh"
#include "G4Sphere.hh"

#include "G4PhysicalConstants.hh"
#include "G4SystemOfUnits.hh"
#include "globals.hh"

#include "G4ThreeVector.hh"
#include "G4MagneticField.hh"


VKDetectorConstruction::VKDetectorConstruction()
{
    fWorld_sol = NULL;
    fWorld_log = NULL;
    fWorld_phys = NULL;

    fMagnet = NULL;
    fBetaDetector = NULL;
    fCatcherFoil = NULL;
    fShield = NULL;
    fLeadBlock = NULL;
    fBlinkers = NULL;
    fFocalPlane = NULL;
    fMidwayPlane = NULL;

    fLabVacuum = NULL;

    fBField.Put(NULL);
    fSiPM_SD.Put(NULL);
    fBetaDetSignal_SD.Put(NULL);
    fBetaDetVeto_SD.Put(NULL);

    fSetSDs = true;

    SetDefaults();

    fDetectorMessenger = new VKDetectorMessenger(this);    
}


VKDetectorConstruction::~VKDetectorConstruction() 
{
    // do nothing
}


G4VPhysicalVolume* VKDetectorConstruction::Construct() 
{
    // construct materials
    fMaterials = VKMaterials::GetInstance();

    // get materials
    fLabVacuum = fMaterials -> GetMaterial("LABVACUUM");

    // construct detector
    return ConstructDetector();
}


G4VPhysicalVolume* VKDetectorConstruction::ConstructDetector()
{
    // clean old geometry
    if (fWorld_phys) 
    {
        G4GeometryManager::GetInstance()->OpenGeometry();
        G4PhysicalVolumeStore::GetInstance()->Clean();
        G4LogicalVolumeStore::GetInstance()->Clean();
        G4SolidStore::GetInstance()->Clean();
        G4LogicalSkinSurface::CleanSurfaceTable();
        G4LogicalBorderSurface::CleanSurfaceTable();
    }

    // Create world volume
    fWorld_sol = new G4Box("world_sol", 1.*m, 1.*m, 1.*m);
    fWorld_log = new G4LogicalVolume(fWorld_sol, fLabVacuum, "world_log", 0, 0, 0);
    fWorld_phys = new G4PVPlacement(0, G4ThreeVector(), fWorld_log, "world", 0, false, 0);
    fWorld_log->SetVisAttributes(G4VisAttributes::Invisible);

    // If the magnet is constructed we place all other objects inside it,
    // otherwise we just place them inside the world volume.
    G4LogicalVolume * mother_log = fWorld_log;

    // Create magnet
    if (fMagnetOn) {
        fMagnet = new VKMagnet(0, G4ThreeVector(), fWorld_log, false, 0, this);
        mother_log = fMagnet -> GetLogicalVolume();
    }
    
    // Create beta detector
    if (fBetaDetectorOn) {
        fBetaDetector = new VKBetaDetector(0, fBetaDetectorPosition, mother_log, false, 0, this);
    }

    // Create catcher foil
    if (fCatcherFoilOn) {
        fCatcherFoil = new VKCatcherFoil(0, fCatcherFoilPosition, mother_log, false, 0, this);
    }
    
    // Create shield
    if (fShieldOn) {
        fShield = new VKShield(0, G4ThreeVector(), mother_log, false, 0, this);
    }

    // Create lead block
    if (fLeadBlockOn) {
        fLeadBlock = new VKLeadBlock(0, G4ThreeVector(), mother_log, false, 0, this);
    }

    // Create blinkers
    if (fBlinkersOn) {
        G4double z = -fSpectrometerLength/2;
        z += fIronEndcapThickness;
        z += fIronConeLength;
        fBlinkers = new VKBlinkers(0, G4ThreeVector(0, 0, z), mother_log, false, 0, this);
    }
    
    // Create virtual detectors to keep track of trajectory
    if (fVirtualDetectorsOn) {
        
        G4String name;
        G4double thickness, rmin, rmax, posZ;
        
        // focal plane
        name      = "focalPlane";
        thickness = 1*mm;
        rmin      = 0;
        rmax      = fBoreDiameter/2;
        posZ      = fBetaDetectorPosition[2] - thickness/2;
        fFocalPlane  = new VKVirtualDetector(
                            0, G4ThreeVector(0, 0, posZ), mother_log, false, 0, this, 
                            name, thickness, rmin, rmax);
        // mid-way plane
        name      = "midwayPlane";
        thickness = 1*mm;
        rmin      = 0;
        if (fShieldOn) rmin = fShieldDiameter/2;
        rmax      = fBoreDiameter/2;
        posZ      = thickness/2;
        fMidwayPlane = new VKVirtualDetector(
                            0, G4ThreeVector(0, 0, posZ), mother_log, false, 0, this, 
                            name, thickness, rmin, rmax);
    }
        
    return fWorld_phys;
}


void VKDetectorConstruction::ConstructSDandField() 
{
    // 
    // Magnetic field
    // 
    if (fBField.Get()==NULL && fBFieldOn)
    {
        // Create magnetic field 
        G4MagneticField* BField =
            new VKTabulatedField3D("res/bfield001.3dmap", fMaxBField);

        // place in cache
        fBField.Put(BField);
          
        // This is thread-local
	    fBFieldMgr = G4TransportationManager::GetTransportationManager() -> GetFieldManager();
               
        fBFieldMgr -> SetDetectorField(fBField.Get());
        fBFieldMgr -> CreateChordFinder(fBField.Get()); 
        // G4cout<< " DeltaStep: " << fBFieldMgr->GetDeltaOneStep()/mm <<" mm\n" << G4endl; 
    }
    else if (!fBFieldOn) 
    {
        fBField.Put(NULL);
        fBFieldMgr -> SetDetectorField(NULL);
        fBFieldMgr -> CreateChordFinder(NULL);     
    }
    

    //
    // Create sensitive detectors only if beta detector is part of setup
    //
    if (fBetaDetectorOn) 
    {
        // SiPMs
        if (!fSiPM_SD.Get()) { // create only once
            // Created here so it exists as SiPMs are being placed
            G4cout << " ---> Constructing SiPM_SD" << G4endl;
            VKSiPMSensitiveDetector* sipm_SD = new VKSiPMSensitiveDetector("SiPM_SD");
            fSiPM_SD.Put(sipm_SD);
            const std::vector<G4ThreeVector> positions = fBetaDetector->GetSiPMPositions();
            sipm_SD -> InitSiPMs(positions.size()); // number of SiPMs
            sipm_SD -> SetSiPMPositions(positions); // their positions
        }
        // Sensitive detector is not actually on the photocathode.
        // processHits gets done manually by the stepping action.
        // It is used to detect when photons hit and get absorbed&detected at the
        // boundary to the photocathode (which doesnt get done by attaching it to a
        // logical volume.
        // It does however need to be attached to something or else it doesnt get
        // reset at the begining of events.
    
        // SIGNAL scintillator detector
        if (!fBetaDetSignal_SD.Get()) { // create only once
            G4cout << " ---> Constructing Signal_SD" << G4endl;
            VKScintillatorSensitiveDetector* signal_SD = new VKScintillatorSensitiveDetector("Signal_SD");
            fBetaDetSignal_SD.Put(signal_SD);
        }

        // VETO scintillator detector
        if (!fBetaDetVeto_SD.Get()) { // create only once
            G4cout << " ---> Constructing Veto_SD" << G4endl;
            VKScintillatorSensitiveDetector* veto_SD = new VKScintillatorSensitiveDetector("Veto_SD");
            fBetaDetVeto_SD.Put(veto_SD);
        }

        if (fSetSDs) {
			// register sensitive detectors with SD manager
			G4SDManager* SDman = G4SDManager::GetSDMpointer();
  			SDman -> AddNewDetector(fSiPM_SD.Get());
  			SDman -> AddNewDetector(fBetaDetSignal_SD.Get());
  			SDman -> AddNewDetector(fBetaDetVeto_SD.Get());
			// register sensitive detectors to logical volumes 
            SetSensitiveDetector(fBetaDetector->GetPhotocathodeLogVol(), fSiPM_SD.Get());
            SetSensitiveDetector(fBetaDetector->GetSignalLogVol(), fBetaDetSignal_SD.Get());
            SetSensitiveDetector(fBetaDetector->GetVetoLogVol(), fBetaDetVeto_SD.Get());
            fSetSDs = false;
        }
    }
}


void VKDetectorConstruction::SetMagnetOn(G4bool b) 
{
    fMagnetOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetBetaDetectorOn(G4bool b) 
{   
    if (!fBetaDetectorOn && b) fSetSDs = true;
    fBetaDetectorOn = b;
//    G4SDManager::GetSDMpointer()->Activate("/VKDet", b); // activate/deactivate sensitive detectors 
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetCatcherFoilOn(G4bool b) 
{
    fCatcherFoilOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetShieldOn(G4bool b) 
{
    fShieldOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetLeadBlockOn(G4bool b) 
{
    fLeadBlockOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetBlinkersOn(G4bool b) 
{
    fBlinkersOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetVirtualDetectorsOn(G4bool b) 
{
    fVirtualDetectorsOn = b;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetBFieldOn(G4bool b) 
{
    fBFieldOn = b;
    // disable/enable magnetic field
    this -> ConstructSDandField();
    // also, reinitialize geometry just to be safe
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetCoatingReflectivity(G4double refl) 
{
    fRefl = refl;
    G4RunManager::GetRunManager()->ReinitializeGeometry();
    G4RunManager::GetRunManager()->Initialize();
}


void VKDetectorConstruction::SetMaxBField(G4double b) 
{
    fMaxBField = b;
    // modify max field strength for current field, if such exists
    VKTabulatedField3D* fPtr = dynamic_cast<VKTabulatedField3D*>(fBField.Get());
    if (fPtr != NULL) fPtr -> SetMaxFieldValue(fMaxBField);
}


void VKDetectorConstruction::SetDefaults() 
{
    // Check for overlaps
    fCheckOverlaps = true;

    // Magnetic field ON/OFF and set max field strength
    fBFieldOn = true;
    fMaxBField = 0.5*tesla;

    // Magnet ON/OFF and bore diameter
    fMagnetOn = true;
    fBoreDiameter = 25*cm;
    fIronEndcapHoleDiameter = 63*mm; 
    fSpectrometerLength     = 80.*cm;
    fIronEndcapThickness    = 4.5*cm;
    fIronConeLength         = 5.5*cm;
    
    // Beta detector ON/OFF and position
    fBetaDetectorOn = true;
    fBetaDetectorPosition = G4ThreeVector(0, 0, 30.0*cm);
    fRefl = 0.99; // from http://dx.doi.org/10.1109/TNS.2008.20032531.0
    
    // Catcher foil ON/OFF and position
    fCatcherFoilOn = true;
    fCatcherFoilPosition = G4ThreeVector(0, 0, -30.0*cm);

    // Shield ON/OFF and diameter
    fShieldOn = true;
    fShieldDiameter = 110*mm;
    
    // Lead block ON/OFF and thickness
    fLeadBlockOn = false;
    
    // Blinkers
    fBlinkersOn = false;
    
    // Virtual detectors ON/OFF
    fVirtualDetectorsOn = true;

    // Re-initialize geometry
    G4RunManager::GetRunManager()->ReinitializeGeometry();
}
