//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKSiPMSensitiveDetector.hh"
#include "VKSiPMHit.hh"
#include "VKDetectorConstruction.hh"
#include "VKUserTrackInformation.hh"

#include "G4VPhysicalVolume.hh"
#include "G4LogicalVolume.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4VTouchable.hh"
#include "G4TouchableHistory.hh"
#include "G4ios.hh"
#include "G4ParticleTypes.hh"
#include "G4ParticleDefinition.hh"


VKSiPMSensitiveDetector::VKSiPMSensitiveDetector(G4String name)
 : G4VSensitiveDetector(name),
   fSiPMHitCollection(0),
   fSiPMPositionsX(0),
   fSiPMPositionsY(0),
   fSiPMPositionsZ(0)
{
    collectionName.insert(name+"_HitCollection");
}


VKSiPMSensitiveDetector::~VKSiPMSensitiveDetector() 
{
    // do nothing
}


void VKSiPMSensitiveDetector::SetSiPMPositions(const std::vector<G4ThreeVector>& positions)
{
    for (G4int i=0; i<G4int(positions.size()); ++i) {
        if (fSiPMPositionsX) fSiPMPositionsX -> push_back(positions[i].x());
        if (fSiPMPositionsY) fSiPMPositionsY -> push_back(positions[i].y());
        if (fSiPMPositionsZ) fSiPMPositionsZ -> push_back(positions[i].z());
    }
}


void VKSiPMSensitiveDetector::Initialize(G4HCofThisEvent* hitsCE)
{
    fSiPMHitCollection = new VKSiPMHitsCollection(SensitiveDetectorName, collectionName[0]);
    // Store collection with event and keep ID
    static G4int hitCID = -1;
    if (hitCID<0) {
        hitCID = GetCollectionID(0);
    }
    hitsCE -> AddHitsCollection(hitCID, fSiPMHitCollection );
}


G4bool VKSiPMSensitiveDetector::ProcessHits(G4Step* ,G4TouchableHistory* )
{
    return false;
}


// Generates a hit and uses the postStepPoint's mother volume replica number
// (PostStepPoint is used because the hit is generated manually when the photon is
// absorbed by the photocathode)

G4bool VKSiPMSensitiveDetector::ProcessHits_constStep(const G4Step* aStep, G4TouchableHistory*)
{

    // Only do something if this is an optical photon
    G4ParticleDefinition * partDef = aStep->GetTrack()->GetDefinition();
    if (partDef != G4OpticalPhoton::OpticalPhotonDefinition()) return false;

    // User replica number 1 since photocathode is a daughter volume          <--- HMMM... IS THIS STILL THE CASE ??!!
    // to the SiPM which was replicated
    G4int SiPMNumber = 
        aStep->GetPostStepPoint()->GetTouchable()->GetReplicaNumber(1);
    G4VPhysicalVolume* physVol = 
        aStep->GetPostStepPoint()->GetTouchable()->GetVolume(1);

    // Loop over existing hits and check if this particular SiPM has 
    // been hit previously in this event
    G4int n = fSiPMHitCollection->entries();
    VKSiPMHit* hit=NULL;
    for (G4int i=0; i<n; i++) {
        if ((*fSiPMHitCollection)[i]->GetSiPMNumber() == SiPMNumber) {
            hit = (*fSiPMHitCollection)[i];
            break;
        }
    }

    // This SiPM wasnt previously hit in this event
    if (hit==NULL) 
    { 
        hit = new VKSiPMHit(); // so create new hit
        hit -> SetSiPMNumber(SiPMNumber);
        hit -> SetSiPMPhysVol(physVol);
        fSiPMHitCollection -> insert(hit);
        double x = (*fSiPMPositionsX)[SiPMNumber];
        double y = (*fSiPMPositionsY)[SiPMNumber];
        double z = (*fSiPMPositionsZ)[SiPMNumber];
        hit -> SetSiPMPos(x, y, z);
    }

    hit -> IncPhotonCount(); // increment photon counter for the selected SiPM

    hit -> SetDrawit(true); // always draw hit

    return true;
}


void VKSiPMSensitiveDetector::EndOfEvent(G4HCofThisEvent* ) 
{
    // do nothing
}


void VKSiPMSensitiveDetector::clear() 
{
    // do nothing
}


void VKSiPMSensitiveDetector::DrawAll() 
{
    // do nothing
}


void VKSiPMSensitiveDetector::PrintAll() 
{
    // do nothing
}


void VKSiPMSensitiveDetector::InitSiPMs(G4int nSiPMs)
{
    if (fSiPMPositionsX) delete fSiPMPositionsX;
    if (fSiPMPositionsY) delete fSiPMPositionsY;
    if (fSiPMPositionsZ) delete fSiPMPositionsZ;
    fSiPMPositionsX = new G4DataVector(nSiPMs);
    fSiPMPositionsY = new G4DataVector(nSiPMs);
    fSiPMPositionsZ = new G4DataVector(nSiPMs);
}
